/*
calc.c
.   __________________________________
.  | _______________________________  |
.  | |    ________________          | |
.  | |   |   ___________  |         | |
.  | |    \  \          |_|         | |
.  | |     \  \                     | |
.  | |      \  \                    | |
.  | |       \  |   Parser          | |
.  | |	    /  /                    | |
.  | |    /  /          _           | |
.  | |   /  /__________| |          | |
.  | |   |_______________|          | |
.  | |  sigmacalc.sourceforge.net   | |
.  | |______________________________| |
.  |__________________________________|
.              /        \
.	          /          \

this file is part of Sigma Calculator
Copyleft (C)2012 SigmaCalc
:~)
http://sigmacalc.sourceforge.net

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include"calc.h"

/*----------------------------------------------------------
Global variables Initializations
----------------------------------------------------------*/

char* scOperatorStrings[]= {"",
	">>","<<",">=","<=","!=","<>","==","&&","||",/* operators with two characters */
	"=","^","%","*","/","\\","-","+","<",">","&","|",/* operators with single character */
};

int scOperatorPrecedence[]= {-1,
	19,19,18,18,17,17,17,14,13,
	1,22,21,21,21,21,20,20,18,18,16,15
};

double (*scFunctions[])(double)= {sin,cos,tan,sinh,cosh,tanh,asin,acos,atan,exp,log,log10,sqrt,ceil,floor,fabs};

char* scFunctionNames[]= {"sin(","cos(","tan(","sinh(","cosh(","tanh(","asin(","acos(","atan(","exp(","log(","log10(","sqrt(","ceil(","floor(","abs("};

char* scErrors[]= {"Unexpected end of line","Syntax error","Parenthesis donot match","Operator not identified","Cannot assign to a constant!"};


/*----------------------------------------------------------
Functions Definitions
----------------------------------------------------------*/
scNode* scNewNode(scNode* Parent)
{
	scNode* New=(scNode*)malloc(sizeof(scNode));
	New->right=0;
	New->parent=Parent;
	New->left=0;
	New->value=0;
	New->Type=0;
	return New;
}

void scDeleteNode(scNode* node)
{
	/* first of all check if this node is a real node or just null value */
	if(!node)
		return;

	if(node->right)
	{
		scDeleteNode(node->right);
	}
	if(node->left)
	{
		scDeleteNode(node->left);
	}
	if(node->Type == SC_VALUE)
		free(node->value);
	free(node);
}

/*
getchar
gets a input from the buffer but doesnot increase the index
*/
char scGetChar()
{
	return scFormula[scPosition];
}

/*
popchar
gets a input from the buffer and increases the index
*/
char scPopChar()
{
	return scFormula[scPosition++];
}

void scSkipSpace()
{
	while(scPopChar()==' ');/* keep popping character until a non-space character is seen */
	scPosition--;
	return;
}

int scEndOfLine()
{
	scSkipSpace();
	return ((scGetChar()=='\0') || (scGetChar()==')') || (scGetChar()==','));
}

scNode* scGetOperator(scNode* Parent)
{
	scNode* Ret;
	int i;
	scSkipSpace();
	Ret=scNewNode(Parent);
	Ret->Type=SC_OPERATOR;

	Ret->value=(void*)-1;
	for(i=1; i<(sizeof(scOperatorStrings)/sizeof(scOperatorStrings[0])); i++)
		if(strstr(scFormula+scPosition,scOperatorStrings[i])==(scFormula+scPosition))
		{
			if(i==SC_ASSIGN && Parent->right->Type != SC_VARIABLE)
			{
				scError.ErrorId=SC_ASSIGN_CONSTANT;
				scError.ErrorPosition=scPosition;
				return Ret;
			}
			scPosition+=strlen(scOperatorStrings[i]);
			Ret->value=(void*)i;
			break;
		}
		if((int)Ret->value==-1)
		{
			scError.ErrorId=SC_SYNTAX_ERROR;
			scError.ErrorPosition=scPosition;
		}
		return Ret;
}

void scPreprocess()
{
	int i,nobracket;
	char temp;
	nobracket=0;//number of brackets
	for(i=0;scFormula[i]!=0;i++)
	{
		temp=scFormula[i];

		if(temp=='{' || temp=='[')
			temp=scFormula[i]='(';
		else if(temp=='}' || temp==']')
			temp=scFormula[i]=')';
		scFormula[i]=tolower(temp);//convert all the text to lower case
		if(temp=='(') nobracket++;
		if(temp==')') nobracket--;
		//if nobracket goes negative, send error
		if(nobracket < 0)
		{
			scError.ErrorId=SC_PARENTHESIS_NOMATCH;
			scError.ErrorPosition=i;
			return;
		}
	}

	//parenthesis not balanced
	if(nobracket != 0)
	{
		scError.ErrorId=SC_PARENTHESIS_NOMATCH;
		scError.ErrorPosition=i+1;
	}
}

scNode* scGetValue(scNode* Parent)
{
	int i;
	scNode* Ret;
	double factor;
	double* dob;

	scSkipSpace();
	if(scGetChar()=='(')//if bracket start here
	{
		scNode* Lt;

		scPosition++;
		Lt=sc_Parse();
		Lt->parent=Parent;

		if(scError.ErrorId>0)//function failed. abort furthur operation
			return Lt;

		if(scGetChar() != ')')
		{
			scError.ErrorId=SC_PARENTHESIS_NOMATCH;
			scError.ErrorPosition=scPosition;
		}
		else
			scPosition++;
		return Lt;
	}
	Ret=scNewNode(Parent);
	if(isalpha(scGetChar()))//if it is alphabet, it must be a function or variable
	{
        //if the next character is not an alphabet then it is a variable
		if(!isalpha(scFormula[scPosition+1] ))
		{
			Ret->Type=SC_VARIABLE;
			Ret->value=(void*)(scPopChar()-'a');
			return Ret;
		}
		for(i=0; i<(sizeof(scFunctionNames)/sizeof(scFunctionNames[0])); i++)
		{
			if(strstr(scFormula+scPosition,scFunctionNames[i])==(scFormula+scPosition))
			{
				scNode* LastNode;
				scPosition+=strlen(scFunctionNames[i]);
				Ret->Type=SC_FUNCTION;
				Ret->value=(void*)i;
				//we extend the function node to the right. Each inner brach represents an argument
				LastNode=Ret;

				LastNode->left=sc_Parse();
				if(scError.ErrorId>0)
					return Ret;
				LastNode=LastNode->left;
				if(scGetChar() != ')')
				{
					scError.ErrorId=SC_PARENTHESIS_NOMATCH;
					scError.ErrorPosition=scPosition;
				}
				else
					scPosition++;
				return Ret;
			}
		}
	}
	Ret->Type=SC_VALUE;
	dob=(double*)malloc(sizeof(double));
	Ret->value=dob;
	*dob=0;
	factor=1.0;//it determines the role of unary operator
	if((scGetChar()=='-') || (scGetChar()=='-'))//unary - or +
	{
		if(scPopChar()=='-')
			factor=-1.0;
	}
	if(!isdigit(scGetChar()) && scGetChar()!='.')
	{

		scError.ErrorId=SC_SYNTAX_ERROR;
		scError.ErrorPosition=scPosition;
		return Ret;
	}

	if(scGetChar()=='0')//binary hexa or octal
	{
		scPopChar();
		if(scGetChar()!='.')
		{
			if(toupper(scGetChar())=='B')//binary
			{
				scPopChar();
				while(scGetChar()=='1' || scGetChar()=='0')
					*dob=(*dob)*2.0 +scPopChar()-'0';
				*dob=(*dob)*factor;
				return Ret;
			}
			else if(toupper(scGetChar())=='X')//hexa decimal
			{
				scPopChar();
				while( (scGetChar()>='0' && scGetChar()<='9') || (toupper(scGetChar())>='A' && toupper(scGetChar())<='F'))
				{
					if(scGetChar()>='0' && scGetChar()<='9')
						*dob=(*dob)*16.0 +scPopChar()-'0';
					else
						*dob=(*dob)*16.0 + toupper(scPopChar())-'A'+10.0;
				}
				*dob=(*dob)*factor;
				return Ret;

			}
			else//else we consider it octal
			{
				while(scGetChar()>='0' && scGetChar()<='7')
					*dob=(*dob)*8.0 +scPopChar()-'0';
				*dob=(*dob)*factor;
				return Ret;
			}
		}

	}
	while(isdigit(scGetChar()))
		*dob=(*dob)*10.0+(double)(scPopChar()-'0');
	if(scGetChar()=='.')
	{
		long a,b;
		scPopChar();
		a=0,b=1;
		while(isdigit(scGetChar()))
		{
			a=a*10+scPopChar()-'0';
			b=b*10;
		}
		//change the long integer to decimal
		*dob=*dob+(double)a/(double)b;
	}
	if(scGetChar()=='e')
	{
		double factor=1;
		long a;
		scPopChar();
		a=0;
		if((scGetChar()=='-') || (scGetChar()=='-'))//unary - or +
		{
			if(scPopChar()=='-')
				factor=-1.0;
		}
		while(isdigit(scGetChar()))
		{
			a=a*10+scPopChar()-'0';
		}
		//change the long integer to decimal
		*dob=(*dob)*pow(10,(double)factor*a);
	}
	*dob=(*dob)*factor;
	return Ret;
}


void scParse()
{
	scHead=sc_Parse();
}

scNode* sc_Parse()
{
	scNode *head,*prev,*temp;
	head=scNewNode(NULL);
	prev=head;

	head->Type=SC_OPERATOR;
	head->value=(void*)0;//initialize it so that it has highest precedence
	//from now on forget that is head...treat it as prev
	prev->right=scGetValue(prev);
	if(scError.ErrorId>0)
		return head;
	//only one number...ie no operators
	if(scEndOfLine())
	{
		temp=prev->right;
		free(head);
		return temp;
	}
	while(!scEndOfLine())
	{
		scNode* Op;
		Op=scGetOperator(prev);
		if(scError.ErrorId>0)
			return head;
		if(scOperatorPrecedence[(int)Op->value] > scOperatorPrecedence[(int)prev->value])
		{
			Op->left=prev->right;
			prev->right=Op;
			Op->parent=prev;
			Op->left->parent=Op;
			Op->right=scGetValue(Op);
			if(scError.ErrorId>0)
				return head;
		}
		else
		{
			while( scOperatorPrecedence[(int)prev->value] > scOperatorPrecedence[(int)Op->value])
				prev=prev->parent;
			Op->left=prev->right;
			Op->parent=prev;
			prev->right->parent=Op;
			prev->right=Op;
			Op->right=scGetValue(Op);
			if(scError.ErrorId>0)
				return head;
		}
		prev=Op;

	}
	temp=head->right;
	free(head);
	return temp;
}


double scEvaluate()
{
	return sc_Evaluate(scHead);
}
double sc_Evaluate(scNode* Head)
{
	if(Head->Type==SC_VALUE)
		return *((double*)Head->value);
	if(Head->Type == SC_VARIABLE)
	{
		return scVariables[(int)Head->value];
	}
	else if(Head->Type == SC_FUNCTION)
	{
			double (*calfunc)(double)=scFunctions[(int)Head->value];
			return (*calfunc)(sc_Evaluate(Head->left));
	}
	else
	{
		switch((int)Head->value)
		{
		case SC_BITWISE_SHIFT_RIGHT:
			{
				return (int)(sc_Evaluate(Head->left)) >> (int)(sc_Evaluate(Head->right));
				break;
			}
		case SC_BITWISE_SHIFT_LEFT:
			{
				return (int)(sc_Evaluate(Head->left)) << (int)(sc_Evaluate(Head->right));
				break;
			}

		case SC_GREATER_OR_EQUAL:
			{
				return ((sc_Evaluate(Head->left)) >= (sc_Evaluate(Head->right)));
				break;
			}
		case SC_SMALLER_OR_EQUAL:
			{
				return ((sc_Evaluate(Head->left)) <= (sc_Evaluate(Head->right)));
				break;
			}
		case SC_NOT_EQUAL1:
		case SC_NOT_EQUAL2:
			{
				return ((sc_Evaluate(Head->left)) != (sc_Evaluate(Head->right)));
				break;
			}
		case SC_EQUAL:
			{
				return ((sc_Evaluate(Head->left)) == (sc_Evaluate(Head->right)));
				break;
			}
		case SC_LOGICAL_AND:
			{
				return ((sc_Evaluate(Head->left)) && (sc_Evaluate(Head->right)));
				break;
			}
		case SC_LOGICAL_OR:
			{
				return ((sc_Evaluate(Head->left)) || (sc_Evaluate(Head->right)));
				break;
			}
		case SC_ASSIGN:
			{
				return (scVariables[(int)(Head->left->value)]=(sc_Evaluate(Head->right)));
				break;
			}
		case SC_POWER:
			{
				return pow(sc_Evaluate(Head->left),sc_Evaluate(Head->right));
				break;
			}
		case SC_MODULUS:
			{
				return (double)((int)sc_Evaluate(Head->left)%(int)sc_Evaluate(Head->right));
				break;
			}
		case SC_MULTIPLICATION:
			{
				return (sc_Evaluate(Head->left)*sc_Evaluate(Head->right));
				break;
			}
		case SC_DIVISION:
			{
				return (sc_Evaluate(Head->left)/sc_Evaluate(Head->right));
				break;
			}
		case SC_INTERGER_DIVISION:
			{
				return (int)(sc_Evaluate(Head->left)/sc_Evaluate(Head->right));
				break;
			}
		case SC_SUBTRACTION:
			{
				return sc_Evaluate(Head->left)-sc_Evaluate(Head->right);
				break;
			}
		case SC_ADDITION:
			{
				return sc_Evaluate(Head->left)+sc_Evaluate(Head->right);
				break;
			}
		case SC_LESS:
			{
				return sc_Evaluate(Head->left)<sc_Evaluate(Head->right);
				break;
			}
		case SC_GREATER:
			{
				return sc_Evaluate(Head->left)>sc_Evaluate(Head->right);
				break;
			}
		case SC_BITWISE_AND:
			{
				return (int)sc_Evaluate(Head->left) & (int)sc_Evaluate(Head->right);
				break;
			}
		case SC_BITWISE_OR:
			{
				return (int)sc_Evaluate(Head->left) | (int)sc_Evaluate(Head->right);
				break;
			}
		};
	}
}
void scSetFormula(const char* Formula)
{
	strcpy(scFormula,Formula);
	scPreprocess();
}
void scReset()
{
	scClear();
	scClearVars();
}
void scClear()
{
	//clear all error message, deletes the node and removes the current formula
	scError.ErrorId = SC_NOERROR;
	scError.ErrorPosition=0;
	strcpy(scFormula,"");
	scPosition=0;
	scDeleteNode(scHead);
	scHead=0;
	return;
}
void scClearVars()
{
	int i;
	for(i=0;i<32;i++)
		scVariables[i]=0.0;
}
double scGetVarValue(char var)
{
	if(var>='a' && var<='z')
		return scVariables[var-'a'];
}
int scSetVarValue(char var,double value)
{
	if(var <'a' || var >'z')
		return 0;
	scVariables[var-'a']=value;
	return 1;
}
