/*
calc.h
.   __________________________________
.  | _______________________________  |
.  | |    ________________          | |
.  | |   |   ___________  |         | |
.  | |    \  \          |_|         | |
.  | |     \  \                     | |
.  | |      \  \                    | |
.  | |       \  |   Parser          | |
.  | |	    /  /                    | |
.  | |    /  /          _           | |
.  | |   /  /__________| |          | |
.  | |   |_______________|          | |
.  | |  sigmacalc.sourceforge.net   | |
.  | |______________________________| |
.  |__________________________________|
.              /        \
.	          /          \

this file is part of Sigma Calculator
Copyleft (C)2012 SigmaCalc
:~)
http://sigmacalc.sourceforge.net

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either
version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#ifndef SIGMA_CALC_H
#define SIGMA_CALC_H

/*----------------------------------------------------------
Header file includes
----------------------------------------------------------*/
#include<stdlib.h>
#include<math.h>
#include<string.h>
#include<ctype.h>
#include<stdio.h>

#pragma function(floor)



/*----------------------------------------------------------
Enumerations declarations
----------------------------------------------------------*/

enum scNodeType {SC_VALUE=1,SC_VARIABLE,SC_OPERATOR,SC_FUNCTION,SC_HEAD};/* Type of the node( or leaf)*/

enum scOperators {SC_DUMMY_OPERATOR=0,

	SC_BITWISE_SHIFT_RIGHT,SC_BITWISE_SHIFT_LEFT,SC_GREATER_OR_EQUAL,SC_SMALLER_OR_EQUAL,
	SC_NOT_EQUAL1,SC_NOT_EQUAL2,SC_EQUAL,SC_LOGICAL_AND,SC_LOGICAL_OR,

	SC_ASSIGN,SC_POWER,SC_MODULUS,SC_MULTIPLICATION,SC_DIVISION,SC_INTERGER_DIVISION,SC_SUBTRACTION,
	SC_ADDITION,SC_LESS,SC_GREATER,SC_BITWISE_AND,SC_BITWISE_OR
};

enum scErrorId {SC_NOERROR=0,SC_UNEXPECTED_END,SC_SYNTAX_ERROR,SC_PARENTHESIS_NOMATCH,SC_OPER_ERROR,SC_ASSIGN_CONSTANT};

/*----------------------------------------------------------
Structure Definitions
----------------------------------------------------------*/
struct _sc_Node
{
	struct _sc_Node *right,*left,*parent;
	int Type;
	void* value;
};

struct
{
	int ErrorId;
	int ErrorPosition;
} scError;

typedef struct _sc_Node scNode;

/*----------------------------------------------------------
Global variables declarations
----------------------------------------------------------*/
#define SC_TOTAL_OPERATORS 22
#define SC_TOTAL_FUNCTIONS 16
int scBufferPointer; /* Pointer to the text buffer */

char* scOperatorStrings[SC_TOTAL_OPERATORS]; /* used to store the operators */
int scOperatorPrecedence[SC_TOTAL_OPERATORS]; /* the precedence of the above operators */

double (*scFunctions[SC_TOTAL_FUNCTIONS])(double); /* array of pointer to the functions */
char* scFunctionNames[SC_TOTAL_FUNCTIONS];
char* scErrors[5];

#ifndef SC_MAX_FORMULA_SIZE
#define SC_MAX_FORMULA_SIZE 1000
#endif
char scFormula[SC_MAX_FORMULA_SIZE];

scNode* scHead;
int scPosition;
double scVariables[32];


/*----------------------------------------------------------
Function declarations
----------------------------------------------------------*/
scNode* scNewNode(scNode* Parent);
void scDeleteNode(scNode* node);
char scGetChar();
char scPopChar();
void scSkipSpace();
int scEndOfLine();
scNode* scGetValue(scNode* Parent);
scNode* scGetOperator(scNode* Parent);
scNode* sc_Parse();
void scParse();
double sc_Evaluate(scNode* Head);
double scEvaluate();

void scSetFormula(const char* Formula);
void scReset();
void scClear();
void scClearVars();
double scGetVarValue(char var);
int scSetVarValue(char var,double value);
void scPreprocess();
#endif
